searchNodes=[{"doc":"An Elixir keyset-based pagination library for Ecto . Why Keyset pagination? The alternative—offset-based pagination—has a couple of serious drawbacks: It's inefficient. The further you &quot;page&quot; into the result set, the less efficient your database queries will be because the database has to query for all the results, then count through and discard rows until it reaches the desired offset. It's inconsistent. Records created or deleted while your user paginates can cause some results to either be duplicated or to be entirely missing from what is returned. Depending on your use case, this could significantly undermine your user's trust or just be awkward. Offset-based pagination is generally simpler to implement because, based on the desired page number and page size, the offset of the next batch can be trivially calculated and incorporated into the next query. With keyset-based pagination, rather than tracking how far into the result set you've traversed (and then hoping records don't change out from under you), we instead track the value of one or more fields in the first and last record of the batch. Then, to get the next (or previous) batch, we query with a WHERE clause that excludes records up to those values. Not only does the database not have to pull irrelevant records only to count through them and eventually discard many of them, but this approach also isn't negatively affected by records being created or removed during pagination. All of this makes keyset pagination far more appealing than offset-based pagination. The gotcha, however, is that it's more troublesome to set up these queries by hand. Fortunately, we've now made it easy to incorporate keyset pagionation into your Elixir/Ecto apps. You're going to love it. For more about the benefits of keyset pagination, see https://use-the-index-luke.com/no-offset. What about Cursor-based pagination? Some people refer to keyset-based pagination as &quot;cursor-based&quot; pagination, which is valid. However, &quot;cursor&quot; pagination does not necessarily imply keyset pagination since offset pagination can also be implemented using cursors. Regardless of keyset vs. offset, cursor-based pagination generally means that the values (either a page number &amp; page size or actual values pulled from a record) are encoded together into an obfuscated form (e.g. Base64 encoded). Use cases Chunkr can help you implement APIs supporting infinite-scroll style interfaces, GraphQL pagination , pagination per the Relay spec , REST APIs, etc. Installation Add chunkr to your list of dependencies in mix.exs : def deps do [ { :chunkr , &quot;~&gt; 0.1.0&quot; } ] end Usage 1. Set up your pagination strategies… defmodule MyApp.PaginatedQueries do use Chunkr.PaginatedQueries paginate_by :account_creation_date do sort :desc , as ( :user ) . inserted_at end paginate_by :profile_name do sort :asc , fragment ( &quot;coalesce(?, &#39;zzzzzzzz&#39;)&quot; , as ( :profile ) . username ) sort :asc , fragment ( &quot;coalesce(?, &#39;zzzzzzzz&#39;)&quot; , as ( :user ) . email_address ) sort :asc , as ( :user ) . id , type : :binary_id end end The paginate_by/1 macro sets up a named pagination strategy, and each call to sort establishes a field to sort by when using this strategy. Results will be ordered by the first specified sort clause, with the next clause acting as a tie-breaker, and so on. Note that the final field provided must be unique in order to provide consistent/deterministic results. Also note that we're coalescing values. Otherwise, any NULL values encountered while filtering against the cursor will simply be dropped and left out of the paginated result set (SQL cannot reasonably compare NULL to an actual value using operators like &lt; and &gt; , so it simply drops them). You'll notice that we must always use Ecto's as clause in order to identify where to find the field in question. This takes advantage of Ecto's late bindings in referencing a query that is yet to be established (you can use many different queries with a single pagination strategy defined here so long as each query provides each of the referenced bindings). The result of registering these pagination strategies is that at compile-time we automatically define functions necessary to take a query and extend it for the desired pagination strategy. This involves dynamically implementing a function to order the results, functions to filter results against any supplied cursor, and a function to automatically retrieve both the records themselves as well as all fields necessary to generate the required cursors. 2. Add chunkr to your Ecto.Repo : defmodule MyApp.Repo do use Ecto.Repo , otp_app : :my_app , adapter : Ecto.Adapters.Postgres use Chunkr , queries : MyApp.PaginatedQueries end This adds the convenience functions paginate/3 and paginate!/3 to your Repo. 3. Paginate your queries! query = from u in User , as : :user , join : ap in assoc ( u , :account_profile ) , as : :profile first_page = MyApp.Repo . paginate! ( query , :profile_name , first : 25 ) next_page = MyApp.Repo . paginate! ( query , :profile_name , first : 25 , after : first_page . end_cursor ) Here we're using the :profile_name pagination strategy established above, and we're providing all of the named bindings required by that strategy (in this case :profile and :user ). See further documentation at Chunkr.PaginatedQueries .","ref":"Chunkr.html","title":"Chunkr","type":"module"},{"doc":"Paginates an Ecto.Queryable . Extends the provided Ecto.Queryable with the necessary filtering, ordering, and cursor field selection for the sake of pagination, then executes the query and returns a Chunkr.Page or results. Opts See Chunkr.Opts","ref":"Chunkr.html#paginate/3","title":"Chunkr.paginate/3","type":"function"},{"doc":"Same as paginate/3 , but raises an error for invalid input.","ref":"Chunkr.html#paginate!/3","title":"Chunkr.paginate!/3","type":"function"},{"doc":"Create and decode opaque, Base64-encoded cursors.","ref":"Chunkr.Cursor.html","title":"Chunkr.Cursor","type":"module"},{"doc":"Decode an opaque cursor.","ref":"Chunkr.Cursor.html#decode/1","title":"Chunkr.Cursor.decode/1","type":"function"},{"doc":"Same as decode/1 but raises an error for invalid cursors.","ref":"Chunkr.Cursor.html#decode!/1","title":"Chunkr.Cursor.decode!/1","type":"function"},{"doc":"Create an opaque, Base64-encoded cursor from cursor_values .","ref":"Chunkr.Cursor.html#encode/1","title":"Chunkr.Cursor.encode/1","type":"function"},{"doc":"","ref":"Chunkr.Cursor.html#t:cursor_values/0","title":"Chunkr.Cursor.cursor_values/0","type":"type"},{"doc":"","ref":"Chunkr.Cursor.html#t:opaque_cursor/0","title":"Chunkr.Cursor.opaque_cursor/0","type":"type"},{"doc":"Options for paginating Fields :query — The non-paginated query to be extended for pagination purposes. :name — The name of the pagination strategy. :cursor — The cursor beyond which results are retrieved. :paging_dir — Either :forward or :backward depending on whether we're paging from the start of the result set toward the end or from the end of the result set toward the beginning. :max_limit — The maximum number of results the user can request per page. :limit — The number of results to actually query for this page. Must be between 0 and max_limit .","ref":"Chunkr.Opts.html","title":"Chunkr.Opts","type":"module"},{"doc":"","ref":"Chunkr.Opts.html#new/3","title":"Chunkr.Opts.new/3","type":"function"},{"doc":"","ref":"Chunkr.Opts.html#t:t/0","title":"Chunkr.Opts.t/0","type":"type"},{"doc":"A single page of results. Fields raw_results — rows in the form {cursor_values, record} where cursor_values is the list of values to be used for generating a cursor. Note that in cases where coalescing or other manipulation was performed for the sake of pagination, the cursor values will reflect that manipulation, while the record itself will by default not. has_previous_page — whether or not there is a previous page of results. has_next_page — whether or not there is a subsequent page of results. start_cursor — a cursor representing the first record in this page of results. end_cursor — a cursor representing the last record in this page of results. opts — opts used.","ref":"Chunkr.Page.html","title":"Chunkr.Page","type":"module"},{"doc":"Returns opaque cursors with their corresponding records.","ref":"Chunkr.Page.html#cursors_and_records/1","title":"Chunkr.Page.cursors_and_records/1","type":"function"},{"doc":"Extracts just the records out of the raw results.","ref":"Chunkr.Page.html#records/1","title":"Chunkr.Page.records/1","type":"function"},{"doc":"Fetches the total, non-paginated count of records that match the query. Counting the total number of records requires an extra database query, so this is not performed by default.","ref":"Chunkr.Page.html#total_count/1","title":"Chunkr.Page.total_count/1","type":"function"},{"doc":"","ref":"Chunkr.Page.html#t:record/0","title":"Chunkr.Page.record/0","type":"type"},{"doc":"","ref":"Chunkr.Page.html#t:t/0","title":"Chunkr.Page.t/0","type":"type"},{"doc":"Provides a set of macros for generating functions to assist with paginating queries. For example: defmodule MyApp.PaginatedQueries do use Chunkr.PaginatedQueries paginate_by :user_created_at do sort :desc , as ( :user ) . inserted_at sort :desc , as ( :user ) . id , type : :binary_id end paginate_by :user_name do sort :asc , fragment ( &quot;lower(coalesce(?, &#39;zzz&#39;)&quot; ) , as ( :user ) . name ) . inserted_at sort :desc , as ( :user ) . id , type : :binary_id end end The paginate_by/1 macro above takes a query name and sets up the necessary beyond_cursor/4 , apply_order/4 , and apply_select/2 functions based on the number of sort options passed in the block as well as the sort directions specified. Each call to sort/3 must include the sort direction, the field to be sorted, and an optional :type keyword. If :type is provided, the cursor value will be cast as that type for the sake of comparisons. See Ecto.Query.API.type/2. Ordering In keyset-based pagination, it is essential that results are deterministically ordered, otherwise you may see unexpected results. Therefore, the final column used for sorting must always be unique and non-NULL. Ordering of paginated results can be based on columns from the primary table, any joined table, any subquery, or any dynamically computed value based on other fields. Regardless of where the column resides, named bindings are always required… Named bindings Because these sort/3 clauses must reference bindings that have not yet been established, each sort clause must use :as to take advantage of late binding. A parallel :as must then be used within the query that gets passed to Chunkr.paginate/3 or the query will fail. See Ecto Named bindings for more. NULL values in sort fields When using comparison operators in SQL, records involving comparisons against NULL get dropped. This is generally undesirable for pagination, as the goal is usually to work your way through an entire result set in chunks—not just through the part of the result set that doesn't have NULL values in the important fields. For example, when sorting users by [last name, first name, middle name], you most likely don't want to exclude users without a known middle name. To work around this awkwardness, you'll need to pick a value that is almost sure to come before or after the rest of your results (depending on whether you want NULL values to sort to the beginning or the end of your results). It's not good enough to think you can simply use a strategy like ordering by NULLS LAST because the filtering of values up to the cursor values will use comparison operators—which will cause records with relevant NULL values to be dropped entirely. The following fragment example sets up names to be compared in a case-insensitive fashion and places records with a NULL name at the end of the list (assuming no names will sort beyond &quot;zzz&quot;!). sort :asc , fragment ( &quot;lower(coalesce(?, &#39;zzz&#39;)&quot; ) , as ( :user ) . name ) . inserted_at Limitations Note that Chunkr limits the number of sort clauses to 4.","ref":"Chunkr.PaginatedQueries.html","title":"Chunkr.PaginatedQueries","type":"module"},{"doc":"Implements the functions necessary for pagination. paginate_by :user_id do sort :asc , as ( :user ) . id end","ref":"Chunkr.PaginatedQueries.html#paginate_by/2","title":"Chunkr.PaginatedQueries.paginate_by/2","type":"macro"}]